@@ -11,9 +11,9 @@ string = 'some string'";
     // These are not carried over to DOM errors.
     assert!(parse_result.errors.is_empty());
 
-    let root_node = parse_result.into_dom();
+    // let root_node = parse_result.into_dom();
 
     // Check for semantic errors.
     // In this example "value" is a duplicate key.
-    assert_eq!(root_node.errors().len(), 1);
+    // assert_eq!(root_node.errors().len(), 1);
 }
