@@ -1,11 +1,10 @@
 //! TOML document to syntax tree parsing.
 
 use crate::{
-    dom,
+    dom::{self, FromSyntax},
     syntax::{SyntaxKind, SyntaxKind::*, SyntaxNode},
     util::{allowed_chars, check_escape},
 };
-use dom::Cast;
 use logos::{Lexer, Logos};
 use rowan::{GreenNode, GreenNodeBuilder, TextRange, TextSize};
 use std::convert::TryInto;
@@ -48,8 +47,10 @@ pub fn parse(source: &str) -> Parse {
 /// A hand-written parser that uses the Logos lexer
 /// to tokenize the source, then constructs
 /// a Rowan green tree from them.
-struct Parser<'p> {
+pub(crate) struct Parser<'p> {
     skip_whitespace: bool,
+    // Allow glob patterns as keys and using [] instead of dots.
+    key_pattern_syntax: bool,
     current_token: Option<SyntaxKind>,
 
     // These tokens are not consumed on errors.
@@ -73,6 +74,22 @@ struct Parser<'p> {
     errors: Vec<Error>,
 }
 
+impl Parser<'_> {
+    /// Required for patch syntax
+    /// and key matches.
+    ///
+    /// It allows a part of glob syntax in identifiers as well.
+    pub(crate) fn parse_key_only(mut self) -> Parse {
+        self.key_pattern_syntax = true;
+        let _ = with_node!(self.builder, KEY, self.parse_key());
+
+        Parse {
+            green_node: self.builder.finish(),
+            errors: self.errors,
+        }
+    }
+}
+
 /// This is just a convenience type during parsing.
 /// It allows using "?", making the code cleaner.
 type ParserResult<T> = Result<T, ()>;
@@ -82,10 +99,11 @@ type ParserResult<T> = Result<T, ()>;
 // this probably has to be rewritten into a state machine
 // that contains minimal function calls.
 impl<'p> Parser<'p> {
-    fn new(source: &'p str) -> Self {
+    pub(crate) fn new(source: &'p str) -> Self {
         Parser {
             current_token: None,
             skip_whitespace: true,
+            key_pattern_syntax: false,
             error_whitelist: 0,
             lexer: SyntaxKind::lexer(source),
             builder: Default::default(),
@@ -104,18 +122,38 @@ impl<'p> Parser<'p> {
 
     fn error(&mut self, message: &str) -> ParserResult<()> {
         let span = self.lexer.span();
-        self.add_error(&Error {
+
+        let err = Error {
             range: TextRange::new(
                 TextSize::from(span.start as u32),
                 TextSize::from(span.end as u32),
             ),
             message: message.into(),
-        });
-        if let Some(t) = self.current_token {
-            if !self.whitelisted(t) {
-                self.token_as(ERROR).ok();
+        };
+
+        let same_error = self
+            .errors
+            .last()
+            .map(|e| e.range == err.range)
+            .unwrap_or(false);
+
+        if !same_error {
+            self.add_error(&Error {
+                range: TextRange::new(
+                    TextSize::from(span.start as u32),
+                    TextSize::from(span.end as u32),
+                ),
+                message: message.into(),
+            });
+            if let Some(t) = self.current_token {
+                if !self.whitelisted(t) {
+                    self.token_as(ERROR).ok();
+                }
             }
+        } else {
+            self.token_as(ERROR).ok();
         }
+
         Err(())
     }
 
@@ -183,6 +221,18 @@ impl<'p> Parser<'p> {
         }
     }
 
+    // This is the same as `token` but won't consume trailing whitespace.
+    fn add_token(&mut self) -> ParserResult<()> {
+        match self.get_token() {
+            Err(_) => Err(()),
+            Ok(token) => {
+                self.builder.token(token.into(), self.lexer.slice());
+                self.current_token = None;
+                Ok(())
+            }
+        }
+    }
+
     fn token(&mut self) -> ParserResult<()> {
         match self.get_token() {
             Err(_) => Err(()),
@@ -190,15 +240,24 @@ impl<'p> Parser<'p> {
         }
     }
 
+    /// This function implicitly calls `step`,
+    /// it was definitely not a good design decision
+    /// but changing this behaviour involves a
+    /// different syntax tree and breakages down the line.
     fn token_as(&mut self, kind: SyntaxKind) -> ParserResult<()> {
+        self.token_as_no_step(kind)?;
+        self.step();
+        Ok(())
+    }
+
+    fn token_as_no_step(&mut self, kind: SyntaxKind) -> ParserResult<()> {
         match self.get_token() {
             Err(_) => return Err(()),
             Ok(_) => {
-                self.builder.token(kind.into(), self.lexer.slice().into());
+                self.builder.token(kind.into(), self.lexer.slice());
             }
         }
 
-        self.step();
         Ok(())
     }
 
@@ -222,18 +281,18 @@ impl<'p> Parser<'p> {
                         }
                     };
 
-                    self.insert_token(token, self.lexer.slice().into());
+                    self.insert_token(token, self.lexer.slice());
                 }
                 WHITESPACE => {
                     if self.skip_whitespace {
-                        self.insert_token(token, self.lexer.slice().into());
+                        self.insert_token(token, self.lexer.slice());
                     } else {
                         self.current_token = Some(token);
                         break;
                     }
                 }
                 ERROR => {
-                    self.insert_token(token, self.lexer.slice().into());
+                    self.insert_token(token, self.lexer.slice());
                     let span = self.lexer.span();
                     self.add_error(&Error {
                         range: TextRange::new(
@@ -339,11 +398,28 @@ impl<'p> Parser<'p> {
     }
 
     fn parse_table_array_header(&mut self) -> ParserResult<()> {
+        self.skip_whitespace = false;
         self.must_token_or(BRACKET_START, r#"expected "[[""#)?;
         self.must_token_or(BRACKET_START, r#"expected "[[""#)?;
+        self.skip_whitespace = true;
         let _ = with_node!(self.builder, KEY, self.parse_key());
+        self.skip_whitespace = false;
         let _ = self.must_token_or(BRACKET_END, r#"expected "]]""#);
-        let _ = self.must_token_or(BRACKET_END, r#"expected "]]""#);
+
+        // Hack in order to avoid calling `step` after
+        // the second closing bracket.
+        let token = self.get_token()?;
+        match token {
+            BRACKET_END => {
+                self.token_as_no_step(token)?;
+            }
+            _ => {
+                self.error(r#"expected "]]"#)?;
+            }
+        }
+        self.skip_whitespace = true;
+
+        self.step();
 
         Ok(())
     }
@@ -358,7 +434,7 @@ impl<'p> Parser<'p> {
 
     fn parse_key(&mut self) -> ParserResult<()> {
         if self.parse_ident().is_err() {
-            return self.error("expected identifier");
+            return self.report_error("expected identifier");
         }
 
         let mut after_period = false;
@@ -382,13 +458,31 @@ impl<'p> Parser<'p> {
                         after_period = true;
                     }
                 }
+                BRACKET_START if self.key_pattern_syntax => {
+                    self.step();
+
+                    match self.parse_ident() {
+                        Ok(_) => {}
+                        Err(_) => return self.error("expected identifier"),
+                    }
+
+                    let token = self.get_token()?;
+
+                    if !matches!(token, BRACKET_END) {
+                        self.error(r#"expected "]""#)?;
+                    }
+                    self.step();
+                    after_period = false;
+                }
                 _ => {
                     if after_period {
                         match self.parse_ident() {
                             Ok(_) => {}
-                            Err(_) => return self.error("expected identifier"),
+                            Err(_) => return self.report_error("expected identifier"),
                         }
                         after_period = false;
+                    } else if self.key_pattern_syntax {
+                        return self.error("unexpected identifier");
                     } else {
                         break;
                     }
@@ -403,6 +497,13 @@ impl<'p> Parser<'p> {
         let t = self.get_token()?;
         match t {
             IDENT => self.token(),
+            IDENT_WITH_GLOB => {
+                if self.key_pattern_syntax {
+                    self.token_as(IDENT)
+                } else {
+                    self.error("expected identifier")
+                }
+            }
             INTEGER_HEX | INTEGER_BIN | INTEGER_OCT => self.token_as(IDENT),
             INTEGER => {
                 if self.lexer.slice().starts_with('+') {
@@ -466,23 +567,23 @@ impl<'p> Parser<'p> {
                 }
             }
             FLOAT => {
-                if self.lexer.slice().starts_with('+') {
+                if self.lexer.slice().starts_with('0') {
+                    self.error("zero-padded numbers are not allowed")
+                } else if self.lexer.slice().starts_with('+') {
                     Err(())
                 } else {
                     for (i, s) in self.lexer.slice().split('.').enumerate() {
                         if i != 0 {
-                            self.insert_token(PERIOD, ".".into());
+                            self.insert_token(PERIOD, ".");
                         }
 
-                        self.insert_token(IDENT, s.into());
+                        self.insert_token(IDENT, s);
                     }
                     self.step();
                     Ok(())
                 }
             }
-            BOOL => {
-                self.token_as(IDENT)
-            }
+            BOOL => self.token_as(IDENT),
             _ => self.error("expected identifier"),
         }
     }
@@ -494,7 +595,31 @@ impl<'p> Parser<'p> {
         };
 
         match t {
-            BOOL | DATE => self.token(),
+            BOOL | DATE_TIME_OFFSET | DATE_TIME_LOCAL | DATE | TIME => self.token(),
+            INTEGER => {
+                // This is probably a logos bug or a priority issue,
+                // for some reason "1979-05-27" gets lexed as INTEGER.
+                if !self.lexer.slice().starts_with('-') && self.lexer.slice().contains('-') {
+                    return self.token_as(DATE);
+                }
+
+                // FIXME: probably another logos bug.
+                if self.lexer.slice().contains(':') {
+                    return self.token_as(TIME);
+                }
+
+                // This could've been done more elegantly probably.
+                if (self.lexer.slice().starts_with('0') && self.lexer.slice() != "0")
+                    || (self.lexer.slice().starts_with("+0") && self.lexer.slice() != "+0")
+                    || (self.lexer.slice().starts_with("-0") && self.lexer.slice() != "-0")
+                {
+                    self.error("zero-padded integers are not allowed")
+                } else if !check_underscores(self.lexer.slice(), 10) {
+                    self.error("invalid underscores")
+                } else {
+                    self.token()
+                }
+            }
             INTEGER_BIN => {
                 if !check_underscores(self.lexer.slice(), 2) {
                     self.error("invalid underscores")
@@ -517,7 +642,23 @@ impl<'p> Parser<'p> {
                 }
             }
             FLOAT => {
-                if !check_underscores(self.lexer.slice(), 10) {
+                // FIXME: probably another logos bug.
+                if self.lexer.slice().contains(':') {
+                    return self.token_as(TIME);
+                }
+
+                let int_slice = if self.lexer.slice().contains('.') {
+                    self.lexer.slice().split('.').next().unwrap()
+                } else {
+                    self.lexer.slice().split('e').next().unwrap()
+                };
+
+                if (int_slice.starts_with('0') && int_slice != "0")
+                    || (int_slice.starts_with("+0") && int_slice != "+0")
+                    || (int_slice.starts_with("-0") && int_slice != "-0")
+                {
+                    self.error("zero-padded numbers are not allowed")
+                } else if !check_underscores(self.lexer.slice(), 10) {
                     self.error("invalid underscores")
                 } else {
                     self.token()
@@ -629,25 +770,16 @@ impl<'p> Parser<'p> {
                     }
                 }
             }
-            INTEGER => {
-                // This could've been done more elegantly probably.
-                if (self.lexer.slice().starts_with('0') && self.lexer.slice() != "0")
-                    || (self.lexer.slice().starts_with("+0") && self.lexer.slice() != "+0")
-                    || (self.lexer.slice().starts_with("-0") && self.lexer.slice() != "-0")
-                {
-                    self.error("zero-padded integers are not allowed")
-                } else if !check_underscores(self.lexer.slice(), 10) {
-                    self.error("invalid underscores")
-                } else {
-                    self.token()
-                }
+            BRACKET_START => {
+                with_node!(self.builder, ARRAY, self.parse_array())
             }
-            BRACKET_START => with_node!(self.builder, ARRAY, self.parse_array()),
-            BRACE_START => with_node!(self.builder, INLINE_TABLE, self.parse_inline_table()),
-            IDENT => {
+            BRACE_START => {
+                with_node!(self.builder, INLINE_TABLE, self.parse_inline_table())
+            }
+            IDENT | BRACE_END => {
                 // FIXME(bit_flags): This branch is just a workaround.
                 self.report_error("expected value").ok();
-                self.token_as(ERROR)
+                Ok(())
             }
             _ => self.error("expected value"),
         }
@@ -659,6 +791,7 @@ impl<'p> Parser<'p> {
         let mut first = true;
         let mut comma_last = false;
         let mut was_newline = false;
+
         loop {
             let t = match self.get_token() {
                 Ok(t) => t,
@@ -673,7 +806,7 @@ impl<'p> Parser<'p> {
                         // table.
                         let _ = self.report_error("expected value, trailing comma is not allowed");
                     }
-                    break self.token()?;
+                    break self.add_token()?;
                 }
                 NEWLINE => {
                     // To avoid infinite loop in case
@@ -686,6 +819,10 @@ impl<'p> Parser<'p> {
                     was_newline = true;
                 }
                 COMMA => {
+                    if comma_last {
+                        let _ = self.report_error(r#"unexpected ",""#);
+                    }
+
                     if first {
                         let _ = self.error(r#"unexpected ",""#);
                     } else {
@@ -719,10 +856,16 @@ impl<'p> Parser<'p> {
         let mut first = true;
         let mut comma_last = false;
         loop {
-            let t = self.get_token()?;
+            let t = match self.get_token() {
+                Ok(t) => t,
+                Err(_) => {
+                    let _ = self.report_error("unexpected EOF");
+                    return Err(());
+                }
+            };
 
             match t {
-                BRACKET_END => break self.token()?,
+                BRACKET_END => break self.add_token()?,
                 NEWLINE => {
                     self.token()?;
                     continue; // as if it wasn't there, so it doesn't count as a first token
@@ -775,7 +918,7 @@ fn check_underscores(s: &str, radix: u32) -> bool {
 
 /// The final results of a parsing.
 /// It contains the green tree, and
-/// the errors that ocurred during parsing.
+/// the errors that occurred during parsing.
 #[derive(Debug, Clone)]
 pub struct Parse {
     pub green_node: GreenNode,
@@ -792,7 +935,7 @@ impl Parse {
     ///
     /// Any semantic errors that occur will be collected
     /// in the returned DOM node.
-    pub fn into_dom(self) -> dom::RootNode {
-        dom::RootNode::cast(rowan::NodeOrToken::Node(self.into_syntax())).unwrap()
+    pub fn into_dom(self) -> dom::node::Node {
+        dom::Node::from_syntax(self.into_syntax().into())
     }
 }
